/* This source file is a part of the GePhex Project.

 Copyright (C) 2001-2004

 Georg Seidel <georg@gephex.org> 
 Martin Bayer <martin@gephex.org> 
 Phillip Promesberger <coma@gephex.org>
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

#include "v4lmodule.h"

#include <stdexcept>
#include <string>
#include <stdint.h>
#include <algorithm>
#include <map>
#include <sstream>

#include "videodevice.h" 
#include "colorconv.h"

static std::map<std::string, VideoDevice*> videoDevices;
static logT s_log;

typedef struct _MyInstance
{
  int videoDeviceDescriptor;
  int blubb;
  int todrop;
} MyInstance, *MyInstancePtr;

int init(logT log_function)
{ 
  s_log = log_function;
  init_tabs();
  return 1;
}


void shutDown(void)
{
}


MyInstance* construct()
{
    MyInstance* my = new MyInstance;
    my->blubb = 0;
    my->todrop=0;
    return my;
}


void destruct(MyInstance* my)
{
  // todo dec some kind of refcount for the last open videodev
  delete my;
}


void update(void* instance)
{
  InstancePtr inst = static_cast<InstancePtr>(instance);
  MyInstancePtr my = inst->my;

  try
    {
      // get the videodevice object by its device name
      std::string deviceFileName(inst->in_devicefile->text);

      std::map<std::string, VideoDevice*>::iterator it =
	videoDevices.find(deviceFileName);
      VideoDevice* videoDev;
      if (it!=videoDevices.end())
	{
	  // the video device is already known
	  // check if it is a valid device
	  if (it->second==0)
	    {
	      // don't retry on every update
	      throw std::runtime_error("no valid device");
	    }
	  videoDev=it->second;
	}
      else
	{
	  // there is no open videodevice 
	  // try to open it, if it fails a exception is thrown
	  try
	    {
	      // this throws an exception if the device could not be opened
	      videoDev = new VideoDevice(deviceFileName,s_log);
	      
	      // add it to the map
	      videoDevices[deviceFileName]=videoDev;
	    }
	  catch(std::runtime_error& e)
	    {
	      // mark it als unvalid
	      videoDevices[deviceFileName]=0;
	      throw;
	    } 
	}
      
      // now we have an open v4L device
      
      // set a/d converter params (v4l1)
      /*
      videoDev->setProperties(
			      trim_int(inst->in_brightness->number*65535.0,
				       0,65535),
			      trim_int(inst->in_hue->number*65535.0,
				       0,65535),
			      trim_int(inst->in_colour->number*65535.0,
				       0,65535),
			      trim_int(inst->in_contrast->number*65535.0,
				       0,65535),
			      trim_int(inst->in_whiteness->number*65535.0,
				       0,65535)
			      );
      */

      // its time to grab a frame with these properties
      VideoDevice::Frame frame;
      frame.xSize=trim_int(inst->in_x_size->number,
			   videoDev->min_width(),videoDev->max_width());
      frame.ySize=trim_int(inst->in_y_size->number,
			   videoDev->min_height(),videoDev->max_height());
  
      // the v4l drivers should not allow modes with undefined pixellayout
      // fix invalid modes here
      frame.xSize-=frame.xSize%2;
      frame.ySize-=frame.ySize%2;
  
      // check if initialized and grab geometry has changed
      if (videoDev->is_initialized() && 
	  (frame.xSize != videoDev->width() ||
	  frame.ySize != videoDev->height()))
	{
	  //log2(2,"geometry has changed");      
	  videoDev->uninitialize();
	}
      
      if (!videoDev->is_initialized())
	{
	  videoDev->initialize(frame);
	}
      
      // get the next image
      videoDev->grabImage(frame, inst->my->todrop != 0);
      
      // update drop counter
      if (inst->my->todrop==0)
	inst->my->todrop=trim_int(inst->in_drops->number,0,200);
      else
	inst->my->todrop--;
      
      // change framebuffer size
      FrameBufferAttributes attribs;
      attribs.xsize=videoDev->width();
      attribs.ysize=videoDev->height();
      framebuffer_changeAttributes(inst->out_r,&attribs);
      
      // convert the grabber framebuffer format to brga32(gephex default)
      switch (frame.type)
	{
	case VIDEO_PALETTE_RGB32: // tested with a bttv 878 card
	  bgra32_2_bgra(reinterpret_cast<uint_8*>(inst->out_r->framebuffer),
			reinterpret_cast<uint_8*>(frame.frameptr),
			frame.xSize,frame.ySize);
	  break;
	case VIDEO_PALETTE_YUV420P: // tested with a phillips webcam
	  // convert the image from yuv411p to bgra
	  yuv411p_2_bgra(reinterpret_cast<uint_8*>(inst->out_r->framebuffer),
			 reinterpret_cast<uint_8*>(frame.frameptr),
			 frame.xSize,frame.ySize);
	  break;
	default:
	  throw std::runtime_error("unsupported pixelformat");
	}
    }

  // in case of any error
  catch (std::runtime_error& e) 
    {
      s_log(2,e.what());
      
      // in case of an error return a dummy image of size 1x1
      FrameBufferAttributes attribs;
      attribs.xsize=1;
      attribs.ysize=1;
      framebuffer_changeAttributes(inst->out_r,&attribs);
      
      // set it to black
      *inst->out_r->framebuffer=0x00000000;
    }
}

