/* This source file is a part of the GePhex Project.

 Copyright (C) 2001-2004

 Georg Seidel <georg@gephex.org> 
 Martin Bayer <martin@gephex.org> 
 Phillip Promesberger <coma@gephex.org>
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

#include "videodevice.h"

#include <sstream>
#include <algorithm>

//#include <sys/types.h>
//#include <sys/stat.h>
//#include <sys/fcntl.h>
//#include <sys/ioctl.h>
//#include <sys/mman.h>
//#include <unistd.h>
//#include <errno.h>
#include <cassert>

#include <iostream>

#include "v4l1.h"

VideoDevice::VideoDevice(const std::string& deviceFileName, logT log2_)   
  throw (std::runtime_error)
  :vd(deviceFileName),
   initialized(false),
   frontbf(1),
   log2(log2_),  
   m_brightness(-1),
   m_hue(-1),
   m_colour(-1),
   m_contrast(-1),
   m_whiteness(-1),
   m_can_select(false),
   m_width(0),
   m_height(0)
{
}

VideoDevice::~VideoDevice()
{
  if (initialized)
    uninitialize();
}

void VideoDevice::uninitialize()
{
  assert(initialized);

  try
    {
      // get all the backbuffers
      vd.vidiocsync(frontbf);

      // and free the alloced memory
      vd.munmap(mmapBase,mmapSize);
    }
  catch (std::runtime_error& e)
    {
      log2(2,e.what());
    }
  
  initialized=false;
}

void VideoDevice::initialize(const Frame& frame)
{
  assert(!initialized);

  //check geometry;
  assert( min_width() <= frame.xSize );
  assert( max_width() >= frame.xSize );
  assert( min_height() <= frame.ySize );
  assert( max_height() >= frame.ySize );
  
  // check capture dimensions
  // try to set the capture geometry(area)
  video_window videoWindow;
  // upper left corner position
  videoWindow.x = 0;
  videoWindow.y = 0;
  // capture area size
  videoWindow.width = frame.xSize;
  videoWindow.height = frame.ySize;
  // options
  videoWindow.flags = 0;
  videoWindow.clipcount = 0;
  videoWindow.clips = NULL;
  
  // set grab dimensions
  vd.vidiocswin(videoWindow);
  

  // get grabdimensions back
  vd.vidiocgwin(videoWindow);
  
  //std::cerr << videoWindow << std::endl;
  
  // and check them
  if (((int)videoWindow.width!=frame.xSize)||
      ((int)videoWindow.height!=frame.ySize))
    {
      std::ostringstream errorMsg;
      errorMsg << "unsupported geometry" << std::endl;
      errorMsg << "requested: " << frame.xSize << " " << frame.ySize;
      throw std::runtime_error(errorMsg.str());
    }
  
  // init memory mapping
  // get mmap info
  video_mbuf videoMBuf;
  
  vd.vidiocgmbuf(videoMBuf);
  
  // check if backbuffering is possible(2 frames)
  if (videoMBuf.frames<2)
    {
      std::string errorMsg;
      errorMsg="mmap cant grab 2 frames";
      throw std::runtime_error(errorMsg);
    }
  
  std::cerr << videoMBuf.frames << std::endl;
  
  // start mmapping the framegrabber memory
  mmapSize=videoMBuf.size;
  
  mmapBase=vd.mmap(mmapSize);
  
  // initialize array with startpointer to the mapped frames
  for(int i=0; i!=2; ++i)
    {
      frameptr[i]=reinterpret_cast<uint_32*>(reinterpret_cast<char*>(mmapBase)+videoMBuf.offsets[i]);
    }
  
  frontbf=0;
  backbf=1;
  
  video_mmap videoMMap; // v4l struct for mmap informations
  videoMMap.frame = frontbf; // buffer position
  videoMMap.width = frame.xSize;
  videoMMap.height = frame.ySize;
  
  // probe pixelformats
  // todo - use a set of pixel format and probe each
  try
    {
      // try RGBA
      videoMMap.format =  VIDEO_PALETTE_RGB32;
      vd.vidiocmcapture(videoMMap);
    }
  catch(std::runtime_error& e)
    {
      try
	{
	  // try YUV
	  videoMMap.format =  VIDEO_PALETTE_YUV420P;
	  vd.vidiocmcapture(videoMMap);
	}
      catch(std::runtime_error& e)
	{
	  throw std::runtime_error("error while starting capture to buffer");
	}
    }
  
  initFrame=frame;
  
  initialized=true;
  m_width=frame.xSize;
  m_height=frame.ySize;
  
  m_pixelformat=videoMMap.format;
}

 
void VideoDevice::grabImage(Frame& frame, bool drop) throw (std::runtime_error)
{
  assert(initialized);

  if (!drop)
    { 
      video_mmap videoMMap; // v4l struct for mmap informations
      videoMMap.frame = backbf; // buffer position
      videoMMap.width = initFrame.xSize;
      videoMMap.height = initFrame.ySize;
      videoMMap.format =  m_pixelformat;
      
      vd.vidiocmcapture(videoMMap);
      std::swap(frontbf,backbf);
      vd.vidiocsync(backbf);
    }
  
  frame.frameptr=frameptr[backbf];
  frame.type=m_pixelformat;
}


void VideoDevice::setProperties(int brightness,int hue,int colour,
				int contrast,
				int whiteness)
{
  video_picture videopic;
  
  vd.vidiocgpict(videopic);
  
  videopic.brightness=brightness;
  videopic.hue=hue;
  videopic.colour=colour;
  videopic.contrast=contrast;
  videopic.whiteness=whiteness;
  
  vd.vidiocspict(videopic);
}

size_t VideoDevice::width()
{
  assert(initialized);
  return m_width;
}

size_t VideoDevice::height()
{
  assert(initialized);
  return m_height;
}

size_t VideoDevice::min_width()
{
  return vd.min_width();
}

size_t VideoDevice::max_width()
{
  return vd.max_width();
}

size_t VideoDevice::min_height()
{
  return vd.min_height();
}

size_t VideoDevice::max_height()
{
  return vd.max_height();
}

bool VideoDevice::is_initialized()
{
  return initialized;
}  
